save_dir                    = '/home/ubuntu/vinc/CloserLookFewShot'
data_dir = {}
data_dir['CUB']             = './filelists/CUB/' 
data_dir['miniImagenet']    = './filelists/miniImagenet/' 
data_dir['omniglot']        = './filelists/omniglot/' 
data_dir['emnist']          = './filelists/emnist/' 
